header {* Testing List Properties *} 

theory 
  List_test 
imports
  List
  Testing
begin


section {* Writing the Test Specification *}
consts is_sorted:: "('a::ord) list \<Rightarrow> bool"
primrec "is_sorted  []      = True"
        "is_sorted  (x#xs)  = ((case xs of [] \<Rightarrow> True 
                                       | y#ys \<Rightarrow> (x < y) \<or> (x = y)) 
                                                             \<and> is_sorted xs)"

section {* Generating test cases *} 
test_spec "is_sorted(PUT (l::('a list)))"
  apply(gen_test_cases 4 1 PUT )
store_test_thm "test_sorting"



section {* Generating test data *}
gen_test_data "test_sorting"

thm test_sorting.test_data
thm test_sorting.test_hyps
thm test_sorting.test_thm

export_test_data "list_data.dat" test_sorting



section {* Test Execution and Result Verification *}

consts_code "op <" ("(_ </ _)")

gen_test_script "list_script.sml" test_sorting PUT "myList.sort"


section {* Improving the Testing Results *}
text {*
Obviously, in reality one would not be satisfied with the test cases generated in 
the section: for testing sorting algorithms one would expect that the test data
somehow represents the set of permutations of the list elements. We have already 
seen, that the test specification used in the last section ``only'' enumerates
lists up to a specific length without any ordering constraints on their
elements. Thus we decide to try a more `'descriptive'' test specification that
is based on the behavior of an insertion sort algorithm:
*} 
consts   ins :: "('a::ord) \<Rightarrow> 'a list \<Rightarrow> 'a list"
primrec "ins x [] = [x]"
        "ins x (y#ys) = (if (x < y) then x#y#ys else (y#(ins x ys)))"

consts   sort:: "('a::ord) list \<Rightarrow> 'a list"
primrec "sort [] = [] "
        "sort (x#xs) = ins x (sort xs)"
text {*
Now we state our test specification by requiring that the behavior of the 
program under test @{text PUT} is identical to the behavior of our 
specified sorting algorithm @{text sort}:
*}
text {*
Based on this specification \verb|gen_test_cases| produces 
test cases representing all permutations of lists up to a fixed
length $n$. Normally, we also want to configure up to which length
lists should be generated (we call this the \emph{depth}
of test case), e.g. if we decide to generated lists up to length
$4$. Our standard setup *}
test_spec "sort l = PUT  l"
  apply(gen_test_cases 4 1 "PUT")
  apply(simp_all)
txt {* 
      which leads to the following test partitioning (excerpt):
      @{subgoals [display, goals_limit=10]}
     *}
  store_test_thm "test_insertion_sort"
text {*
generates 34 test cases describing all permutations of lists of length
$1,2,3$ and $4$.
*}

text {*
Generating concrete test data already takes a remarkable length of
time, as it's quite unlikely that the random solver generates values
that fulfills these ordering constraints.  Therefore we restrict the
attempts (\emph{iterations}) the random solver takes for solving a
single test case to $10$.
*}
testgen_params [iterations=10]
gen_test_data "test_insertion_sort"

thm  test_insertion_sort.test_data
text {*
which is sadly not sufficient to solve all conditions, e.g. we 
obtain test cases like:
@{thm [display] test_insertion_sort.test_data}
*}
text {*
were @{text RSF} marks unsolved cases. Analyzing the generated
test data reveals, that all cases for lists with length up to (and including) 
3 could be solved. From the 24 cases for lists of length 4 only
9 could be solved by the random solver 
(thus, overall 19 of the 34 cases 
were solved). To achieve more concrete test cased, we could 
interactive increase the number of iterations which reveals that we
need to set iterations to 100 to find all solutions reliably:r 
\begin{center}
 \begin{tabular}{l||r|r|r|r|r|r|r|r|r}
 iterations&  5 & 10 & 20 & 25 & 30 & 40 & 50 & 75 & 100 \\
 \hline
 solved goals (of 34) & 13 & 19 & 23 & 24 & 25 & 29 & 33 & 33 &  34      
 \end{tabular}
\end{center}
Instead of increasing the number of iterations one could 
also add  other techniques such as 
        \begin{enumerate}
        \item deriving new rules that allow for 
              the generation of a simplified test theorem,
        \item introducing abstract test cases or 
        \item supporting the solving process by derived rules.
        \end{enumerate}

      *}

end
